#ifndef BLD_RULE_H
#define BLD_RULE_H

/******************************************************************************

Copyright (c) 1997, Andrew J. Weber, Jason P. Weber

Copyright information should have been included with this distribution
and may be found at ../meta/COPYRIGHT or obtained via email from weber@acm.org

******************************************************************************/

#include <gpl/gpl.h>
#include <bld/op.h>

/***********************************************************************
enum BLD_RuleType

Type of rule.
***********************************************************************/
enum BLD_RuleType
{
	BLD_RULE_FAIL_OPERATIONS		=	-2,
	BLD_RULE_PREPEND_OPERATIONS		=	-1,
	BLD_RULE_OVERWRITE_OPERATIONS	=	0,
	BLD_RULE_APPEND_OPERATIONS		=	1
};

/***********************************************************************
class BLD_RuleBase

Rule base.
***********************************************************************/
class BLD_RuleBase
{
	public:
		BLD_RuleBase(void) : type(BLD_RULE_OVERWRITE_OPERATIONS) {}
		BLD_RuleBase(const BLD_RuleType t, const OpArray & o)
			: type(t), operations(o) {}
		virtual ~BLD_RuleBase(void) {}

		inline BLD_RuleType GetType(void) const { return type; }
		inline const OpArray & GetOperations(void) const { return operations; }

		virtual GPL_Boolean Matches(const GPL_String & string,
									GPL_Array<GPL_String> & expressions)
		{
			expressions.Clear();
			return FALSE;
		}

		static GPL_Boolean IsVariable(const GPL_String & string);
		static GPL_Boolean IsPattern(const GPL_String & string);

	protected:
		BLD_RuleType type;
		OpArray operations;
};

/***********************************************************************
class BLD_Rule

Rule whose target specification is a constant string
***********************************************************************/
class BLD_Rule : public BLD_RuleBase
{
	public:
		BLD_Rule(void) {}
		BLD_Rule(const BLD_RuleType t, const OpArray & o)
			: BLD_RuleBase(t, o) {}
		BLD_Rule(const GPL_String & s, const BLD_RuleType t, const OpArray & o)
			: BLD_RuleBase(t, o), target(s) {}
		virtual ~BLD_Rule(void) {}

inline GPL_Boolean operator==(const BLD_Rule & rule) const { return TRUE; }
		inline GPL_String GetTarget(void) const { return target; }

		virtual GPL_Boolean Matches(const GPL_String & string,
									GPL_Array<GPL_String> & expressions);

	protected:
		GPL_String target;
};

/***********************************************************************
class BLD_VariableRule

Rule whose target specification is a variable string
***********************************************************************/
class BLD_VariableRule : public BLD_Rule
{
	public:
		BLD_VariableRule(void) {}
		BLD_VariableRule(const BLD_RuleType t, const OpArray & o)
			: BLD_Rule(t, o) {}
		BLD_VariableRule(const GPL_String & s, const BLD_RuleType t,
							const OpArray & o) : BLD_Rule(s, t, o) {}
		virtual ~BLD_VariableRule(void) {}

		virtual GPL_Boolean Matches(const GPL_String & string,
									GPL_Array<GPL_String> & expressions);
};

/***********************************************************************
class BLD_PatternRule

Rule whose target specification is a constant pattern
***********************************************************************/
class BLD_PatternRule : public BLD_RuleBase
{
	public:
		BLD_PatternRule(void) {}
		BLD_PatternRule(const BLD_RuleType t, const OpArray & o)
			: BLD_RuleBase(t, o) {}
		BLD_PatternRule(const GPL_String & s, const BLD_RuleType t,
						const OpArray & o)
			: BLD_RuleBase(t, o), regexp(s) {}
		virtual ~BLD_PatternRule(void) {}

		virtual GPL_Boolean Matches(const GPL_String & string,
									GPL_Array<GPL_String> & expressions);

	protected:
		GPL_RegularExpression regexp;
};

/***********************************************************************
class BLD_VariablePatternRule

Rule whose target specification is a constant pattern
***********************************************************************/
class BLD_VariablePatternRule : public BLD_PatternRule
{
	public:
		BLD_VariablePatternRule(void) {}
		BLD_VariablePatternRule(const BLD_RuleType t, const OpArray & o)
			: BLD_PatternRule(t, o) {}
		BLD_VariablePatternRule(const GPL_String & s, const BLD_RuleType t,
						const OpArray & o)
			: BLD_PatternRule(s, t, o), pattern(s) {}
		virtual ~BLD_VariablePatternRule(void) {}

		virtual GPL_Boolean Matches(const GPL_String & string,
									GPL_Array<GPL_String> & expressions);

	protected:
		GPL_String pattern;
};

#endif /* BLD_RULE_H */

