/******************************************************************************

Coral Tree dna/binding.cc
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include <gpl/gpl.h>
#include <wba/wba.h>
#include <dna/dna.h>

/******************************************************************************
class DNA_Binding
******************************************************************************/

/***********************************************************************
void DNA_Binding::Detach(void)
***********************************************************************/
void DNA_Binding::Detach(void)
{
	// If this is a valid reference and it is the last one to this
	// implementation (presuming that the user has not sent the same
	// implementation to more than one binding with SetImp())...
	if (GPL_Reference::IsValid() && !GPL_Reference::IsShared())
	{
		// Close the binding.
		// NOTE: No unbinding is performed.  This follows the
		// UDP/TCP model; its genericity has yet to be verified.
		if (DNA_CHECK_STATE(DNA_BINDING_OPEN))
		{
			Close();
		}
	}
	GPL_Reference::Detach();
}

/***********************************************************************
GPL_Boolean DNA_Binding::IsOpen(void) const
***********************************************************************/
GPL_Boolean DNA_Binding::IsOpen(void) const
{
	register DNA_BindingRep *r = (DNA_BindingRep *) rep;
	if ((DNA_BindingRep *) GPL_BAD_ADDRESS != r)
	{
		if (DNA_BINDING_OPEN ==
			(r->state & DNA_BINDING_OPEN))
		{
			return TRUE;
		}
	}
	return FALSE;
}

/***********************************************************************
GPL_Boolean DNA_Binding::IsBound(void) const
***********************************************************************/
GPL_Boolean DNA_Binding::IsBound(void) const
{
	register DNA_BindingRep *r = (DNA_BindingRep *) rep;
	if ((DNA_BindingRep *) GPL_BAD_ADDRESS != r)
	{
		if (DNA_BINDING_BOUND ==
			(r->state & DNA_BINDING_BOUND))
		{
			return TRUE;
		}
	}
	return FALSE;
}

/***********************************************************************
DNA_Point *DNA_Binding::GetSource(void) const
***********************************************************************/
DNA_Point *DNA_Binding::GetSource(void) const
{
	if (GPL_Reference::IsValid())
	{
		register DNA_BindingRep *r = (DNA_BindingRep *) rep;
		if ((DNA_Point *) GPL_BAD_ADDRESS != r->source)
		{
			wbaLog("DNA LOG", "GetSource: %s", r->source->ToString().GetData());
		}
		return r->source;
	}
	return (DNA_Point *) GPL_BAD_ADDRESS;
}

/***********************************************************************
DNA_Point *DNA_Binding::GetDestination(void) const
***********************************************************************/
DNA_Point *DNA_Binding::GetDestination(void) const
{
	if (GPL_Reference::IsValid())
	{
		register DNA_BindingRep *r = (DNA_BindingRep *) rep;
		if ((DNA_Point *) GPL_BAD_ADDRESS != r->destination)
		{
			wbaLog("DNA LOG", "GetDestination: %s", r->destination->ToString().GetData());
		}
		return r->destination;
	}
	return (DNA_Point *) GPL_BAD_ADDRESS;
}

/***********************************************************************
void DNA_Binding::SetImp(DNA_BusImp *i)
***********************************************************************/
void DNA_Binding::SetImp(DNA_BusImp *i)
{
	GPL_ENSURE(DNA_BindingRep, r);r=r;
	DNA_Bus::SetImp(i);
}

/***********************************************************************
void DNA_Binding::SetSource(DNA_Point *p)
***********************************************************************/
void DNA_Binding::SetSource(DNA_Point *p)
{
	GPL_ENSURE(DNA_BindingRep, r);
	if ((DNA_Point *) 0 != p)
	{
		wbaLog("DNA LOG", "SetSource: %s", p->ToString().GetData());
	}
	r->source = p;
}

/***********************************************************************
void DNA_Binding::SetDestination(DNA_Point *p)
***********************************************************************/
void DNA_Binding::SetDestination(DNA_Point *p)
{
	GPL_ENSURE(DNA_BindingRep, r);
	if ((DNA_Point *) 0 != p)
	{
		wbaLog("DNA LOG", "SetDestination: %s", p->ToString().GetData());
	}
	r->destination = p;
	if (!DNA_NO_IMP)
	{
		((DNA_BindingImp *) r->imp)->SetDestination(p);
	}
}

/***********************************************************************
void DNA_Binding::Reset(void)
***********************************************************************/
void DNA_Binding::Reset(void)
{
	DNA_Bus::Reset();
	register DNA_BindingRep *r = (DNA_BindingRep *) rep;
	if ((DNA_BindingRep *) GPL_BAD_ADDRESS != r)
	{
		r->source = (DNA_Point *) GPL_BAD_ADDRESS;
		r->destination = (DNA_Point *) GPL_BAD_ADDRESS;
	}
}

/***********************************************************************
GPL_Boolean DNA_Binding::Open(void)
***********************************************************************/
GPL_Boolean DNA_Binding::Open(void)
{
	GPL_ENSURE(DNA_BindingRep, r);
	if (!DNA_CHECK_STATE(DNA_BINDING_VALID))
	{
		error.Set(DNA_ERROR_BAD_STATE);
		DNA_LOG_ERROR;
		return FALSE;
	}

	if (DNA_NO_IMP)
	{
		error.Set(DNA_ERROR_BAD_IMPLEMENTATION);
		DNA_LOG_ERROR;
		return FALSE;
	}

	if (!((DNA_BindingImp *) r->imp)->Open(error))
	{
		if (!DNA_CHECK_STATE(DNA_BINDING_VALID))
		{
			DNA_CLEAR_STATE(DNA_BINDING_OPEN);
		}
		switch(error.GetType())
		{
		case DNA_ERROR_INSUFFICIENT_MEMORY:
			// Falls through...
		case DNA_ERROR_NO_PERMISSION:
			// Falls through...
		case DNA_ERROR_INSUFFICIENT_RESOURCE:
			// Falls through...
		case DNA_ERROR_BAD_PROTOCOL:
			DNA_CLEAR_STATE(DNA_BINDING_OPEN);
			DNA_LOG_ERROR;
			break;
		default:
			// NOP
			break;
		}
		return FALSE;
	}

	DNA_SET_STATE(DNA_BINDING_OPEN);
	return TRUE;
}

/***********************************************************************
GPL_Boolean DNA_Binding::Bind(void)
***********************************************************************/
GPL_Boolean DNA_Binding::Bind(void)
{
	GPL_ENSURE(DNA_BindingRep, r);
	if (!DNA_CHECK_STATE(DNA_BINDING_VALID | DNA_BINDING_OPEN))
	{
		error.Set(DNA_ERROR_BAD_STATE);
		DNA_LOG_ERROR;
		return FALSE;
	}

	if (DNA_NO_IMP)
	{
		error.Set(DNA_ERROR_BAD_IMPLEMENTATION);
		DNA_LOG_ERROR;
		return FALSE;
	}

	if (!((DNA_BindingImp *) r->imp)->Bind(r->source, error))
	{
		if (!DNA_CHECK_STATE(DNA_BINDING_VALID))
		{
			DNA_CLEAR_STATE(DNA_BINDING_OPEN | DNA_BINDING_BOUND);
		}
		switch(error.GetType())
		{
		case DNA_ERROR_BAD_HANDLE:
			// Falls through...
		case DNA_ERROR_IO_ERROR:
			// Falls through...
		case DNA_ERROR_NO_PERMISSION:
			// Falls through...
		case DNA_ERROR_ADDRESS_IN_USE:
			// Falls through...
		case DNA_ERROR_ADDRESS_UNAVAILABLE:
			// Falls through...
		case DNA_ERROR_INSUFFICIENT_RESOURCE:
			DNA_CLEAR_STATE(DNA_BINDING_OPEN | DNA_BINDING_BOUND);
			DNA_LOG_ERROR;
			break;
		default:
			// NOP
			break;
		}
		return FALSE;
	}

	DNA_SET_STATE(DNA_BINDING_BOUND);
	return TRUE;
}

/***********************************************************************
GPL_Boolean DNA_Binding::Close(void)
***********************************************************************/
GPL_Boolean DNA_Binding::Close(void)
{
	GPL_ENSURE(DNA_BindingRep, r);
	if (!DNA_CHECK_STATE(DNA_BINDING_VALID))
	{
		error.Set(DNA_ERROR_BAD_STATE);
		DNA_LOG_ERROR;
		return FALSE;
	}

	if (DNA_NO_IMP)
	{
		error.Set(DNA_ERROR_BAD_IMPLEMENTATION);
		DNA_LOG_ERROR;
		return FALSE;
	}

	if (!((DNA_BindingImp *) r->imp)->Close(error))
	{
		switch(error.GetType())
		{
		case DNA_ERROR_INSUFFICIENT_RESOURCE:
			// Falls through...
		case DNA_ERROR_BAD_PROTOCOL:
			// Falls through...
		case DNA_ERROR_NO_PERMISSION:
			// Falls through...
		case DNA_ERROR_ADDRESS_IN_USE:
			// Falls through...
		case DNA_ERROR_ADDRESS_UNAVAILABLE:
			// Falls through...
		default:
			DNA_LOG_ERROR;
			break;
		}
		return FALSE;
	}

		DNA_CLEAR_STATE(DNA_BINDING_OPEN | DNA_BINDING_BOUND);
		return TRUE;
}

/***********************************************************************
GPL_Boolean DNA_Binding::Read(void *buffer, const unsigned long size,
					unsigned long & length)
***********************************************************************/
GPL_Boolean DNA_Binding::Read(void *buffer, const unsigned long size,
					unsigned long & length)
{
	GPL_ENSURE(DNA_BindingRep, r);
	if (!DNA_CHECK_STATE(DNA_BINDING_OPEN | DNA_BINDING_BOUND))
	{
		error.Set(DNA_ERROR_BAD_STATE);
		DNA_LOG_ERROR;
		return FALSE;
	}
	register GPL_Boolean success = DNA_Bus::Read(buffer, size, length);
	if (!success)
	{
		if (!DNA_CHECK_STATE(DNA_BINDING_VALID))
		{
			DNA_CLEAR_STATE(DNA_BINDING_OPEN | DNA_BINDING_OPEN);
		}
	}
	if (!DNA_NO_IMP)
	{
		r->destination = ((DNA_BindingImp *) r->imp)->GetDestination();
	}

	return success;
}

/***********************************************************************
GPL_Boolean DNA_Binding::Read(void *buffer, const unsigned long size,
							const long timeout, unsigned long & length)
***********************************************************************/
GPL_Boolean DNA_Binding::Read(void *buffer, const unsigned long size,
							const long timeout, unsigned long & length)
{
	GPL_ENSURE(DNA_BindingRep, r);
	if (!DNA_CHECK_STATE(DNA_BINDING_OPEN | DNA_BINDING_BOUND))
	{
		error.Set(DNA_ERROR_BAD_STATE);
		DNA_LOG_ERROR;
		return FALSE;
	}
	register GPL_Boolean success = DNA_Bus::Read(buffer, size, timeout, length);
	if (!success)
	{
		if (!DNA_CHECK_STATE(DNA_BINDING_VALID))
		{
			DNA_CLEAR_STATE(DNA_BINDING_OPEN | DNA_BINDING_OPEN);
		}
	}
	if (!DNA_NO_IMP)
	{
		r->destination = ((DNA_BindingImp *) r->imp)->GetDestination();
	}

	return success;
}

/***********************************************************************
GPL_Boolean DNA_Binding::Write(void *buffer, const unsigned long size,
					unsigned long & length)
***********************************************************************/
GPL_Boolean DNA_Binding::Write(void *buffer, const unsigned long size,
					unsigned long & length)
{
	GPL_ENSURE(DNA_BindingRep, r);
	if (!DNA_CHECK_STATE(DNA_BINDING_OPEN | DNA_BINDING_BOUND))
	{
		error.Set(DNA_ERROR_BAD_STATE);
		DNA_LOG_ERROR;
		return FALSE;
	}
	if (!DNA_NO_IMP)
	{
		((DNA_BindingImp *) r->imp)->SetDestination(r->destination);
	}
	if (!DNA_Bus::Write(buffer, size, length))
	{
		if (!DNA_CHECK_STATE(DNA_BINDING_VALID))
		{
			DNA_CLEAR_STATE(DNA_BINDING_OPEN | DNA_BINDING_OPEN);
		}
		switch(error.GetType())
		{
		case DNA_ERROR_CONNECTION_RESET:
			DNA_CLEAR_STATE(DNA_BINDING_OPEN | DNA_BINDING_OPEN);
			DNA_LOG_ERROR;
			break;
		default:
			// NOP
			break;
		}
		return FALSE;
	}
	return TRUE;
}

/***********************************************************************
GPL_Boolean DNA_Binding::Print(GPL_String & string,
							const GPL_Verbosity verbosity) const
***********************************************************************/
GPL_Boolean DNA_Binding::Print(GPL_String & string,
							const GPL_Verbosity verbosity) const
{
	if (GPL_Reference::IsValid())
	{
		register DNA_BindingRep *r = (DNA_BindingRep *) rep;
		GPL_String s;
		if ((DNA_BindingImp *) GPL_BAD_ADDRESS == r->imp)
		{
			string.Append("null");
		}
		else
		{
			s.Printf("%s(%p)", r->imp->GetClassName(), (void *) r->imp);
			string.Append(s);
		}
		string.Append(',');
		s.Clear();
		if (r->state & DNA_BINDING_VALID)
		{
			s.Join("|", "VALID");
		}
		if (r->state & DNA_BINDING_OPEN)
		{
			s.Join("|", "OPEN");
		}
		if (r->state & DNA_BINDING_BOUND)
		{
			s.Join("|", "BOUND");
		}
		string.Append(s);
	}
	return TRUE;
}

GPL_OBJECT_IMPLEMENTATION(DNA_Binding, 0x40070000, "DNA_Binding");

/***********************************************************************
class DNA_BindingRep
***********************************************************************/
DNA_BindingRep::~DNA_BindingRep(void)
{
	if ((DNA_Point *) GPL_BAD_ADDRESS != source)
	{
		delete source;
	}
	if ((DNA_Point *) GPL_BAD_ADDRESS != destination)
	{
		delete destination;
	}
}
