/******************************************************************************

Coral Tree dna/xdnaclnt.cc
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include <gpl/gpl.h>
#include <wba/wba.h>
#include <dna/dna.h>

enum DNA_TestOperation
{
	DNA_TEST_NOP			= 0x00,
	DNA_TEST_PING		= 0x01,
	DNA_TEST_SHUTDOWN	= 0xFF
};

#define DNA_DECLARE_HANDLER(className, name, usage, description)	\
class className : public WBA_CommandHandler	\
{	\
	public:	\
		className(void) { names.AppendElement( name ); }	\
		~className(void) {} \
		const char *GetUsage(void) const { return usage ; }	\
		const char *GetDescription(void) const { return description ; }	\
		GPL_Boolean HandleCommand(void);	\
}

/***********************************************************************
class DNA_TestClient

***********************************************************************/
class DNA_TestClient : public WBA_Shell, public DNA_Client
{
	public:
		DNA_TestClient(void);
		virtual ~DNA_TestClient(void);

		virtual const char * GetName(void) const;
		virtual const char * GetParameterUsage(void) const;

		virtual GPL_Boolean Initialize(void);
		virtual void Finalize(void);

		void RequestShutdown(void);
		void Ping(long size);
		virtual GPL_Boolean Reconnect(void);

	protected:
		DNA_DECLARE_HANDLER(PingHandler, "ping", "", "Ping the test server");
		DNA_DECLARE_HANDLER(ReconnectHandler, "reconnect", "", "Reconnect to the test server");
		DNA_DECLARE_HANDLER(ShutdownHandler, "shutdown", "", "Shutdown the test server");

		virtual void Process(void);
		virtual void ProcessMessage(DNA_Message & m);
	private:
		long pingSendTime;
};

/***********************************************************************
DNA_TestClient::DNA_TestClient
***********************************************************************/
DNA_TestClient::DNA_TestClient(void) : WBA_Shell(this)
{
}

/***********************************************************************
DNA_TestClient::~DNA_TestClient
***********************************************************************/
DNA_TestClient::~DNA_TestClient(void)
{
}

/******************************************************************************
const char * DNA_TestClient::GetName(void) const
******************************************************************************/
const char * DNA_TestClient::GetName(void) const
{
	static const char name[] = "DNA_TestClient";
	return name;
}

/******************************************************************************
const char * DNA_TestClient::GetParameterUsage(void) const
******************************************************************************/
const char * DNA_TestClient::GetParameterUsage(void) const
{
	static const char parameters[] = "";
	return parameters;
}

/***********************************************************************
GPL_Boolean DNA_TestClient::Initialize(void)
***********************************************************************/
GPL_Boolean DNA_TestClient::Initialize(void)
{
	// Set the prompt.
	SetPrompt("DNA_TestClient> ");

	// Add the commands.
	AddCommand(new WBA_CommandHelp());
	AddCommand(new WBA_CommandQuit());
	AddCommand(new PingHandler());
	AddCommand(new ReconnectHandler());
	AddCommand(new ShutdownHandler());

	// Set-up the destination address.
	DNA_IPPoint *destination = new DNA_IPPoint();
	destination->SetHost(WBA_ARGS.Find("DNA_TEST_SERVER")->GetValue());
	destination->SetPort((unsigned short)
						(0xFFFF & WBA_ARGS.Find("DNA_TEST_PORT")->GetLong()));
	SetDestination(destination);

	DNA_Client::Initialize();
	if (!IsConnected())
	{
		WBA_APP->Printf("Unable to connect to server on %s",
						WBA_ARGS.Find("DNA_TEST_SERVER")->GetValue().GetData());
		SetExitCode(-1);
		Exit();
	}

	// Wait for some time period for responses.
	DNA_Client::SetTimeout(500);

	return WBA_Shell::Initialize();
}

/***********************************************************************
void DNA_TestClient::Finalize(void)
***********************************************************************/
void DNA_TestClient::Finalize(void)
{
	DNA_Client::Finalize();
	WBA_Shell::Finalize();
}

/***********************************************************************
void DNA_TestClient::RequestShutdown(void);
***********************************************************************/
void DNA_TestClient::RequestShutdown(void)
{
	DNA_Message m;
	m.AppendUnsignedChar((unsigned char) DNA_TEST_SHUTDOWN);
	Request(m);
}

/***********************************************************************
void DNA_TestClient::Ping(void);
***********************************************************************/
void DNA_TestClient::Ping(long size)
{
	DNA_Message m;
	m.AppendUnsignedChar((unsigned char) DNA_TEST_PING);
	if(size)
	{
		char *block;
		block = (char *)osdMalloc(size);
		m.AppendBlock(block, size);
		osdFree(block);
	}
	pingSendTime = osdTimer();
	if (!Request(m))
	{
		WBA_APP->Printf("Unable to ping server\n");
	}
}

/***********************************************************************
GPL_Boolean DNA_TestClient::Reconnect(void);
***********************************************************************/
GPL_Boolean DNA_TestClient::Reconnect(void)
{
	DNA_Client::Finalize();
	return DNA_Client::Reconnect();
}

/***********************************************************************
void DNA_TestClient::Process(void)
***********************************************************************/
void DNA_TestClient::Process(void)
{
	DNA_Client::Process();
}

/***********************************************************************
void DNA_TestClient::ProcessMessage(DNA_Message & message)
***********************************************************************/
void DNA_TestClient::ProcessMessage(DNA_Message & m)
{
	register const DNA_TestOperation op =
						(DNA_TestOperation) m.ReadUnsignedChar();
WBA_APP->Printf("Processing operation %X\n", (long) op);
	switch (op)
	{
	case DNA_TEST_NOP:
		// NOP
		break;
	case DNA_TEST_PING:
		if (TRUE == (GPL_Boolean) m.ReadUnsignedChar())
		{
			WBA_APP->Printf("Server is alive. Roundtrip time: %d\n", 
				osdTimer() - pingSendTime);
		}
		else
		{
			WBA_APP->Printf("Server is dead!\n");
		}
		break;
	case DNA_TEST_SHUTDOWN:
		WBA_APP->Printf("Nonsensical operation %X", op);
		break;
	default:
		WBA_APP->Printf("Unknown operation %X", op);
		break;
	}
}

/***********************************************************************
GPL_Boolean DNA_TestClient::PingHandler::HandleCommand(void)
***********************************************************************/
GPL_Boolean DNA_TestClient::PingHandler::HandleCommand(void)
{
	long size = 0;
	if(arguments.GetCount() > 1)
	{
		size = atol((const char *)(arguments.GetElement(1)));
	}
	((DNA_TestClient *) WBA_Application::GetCurrentApplication())->Ping(size);
	return TRUE;
}

/***********************************************************************
GPL_Boolean DNA_TestClient::ReconnectHandler::HandleCommand(void)
***********************************************************************/
GPL_Boolean DNA_TestClient::ReconnectHandler::HandleCommand(void)
{
	if (!((DNA_TestClient *) WBA_Application::GetCurrentApplication())->Reconnect())
	{
		WBA_APP->Printf("Unable to reconnect to server");
	}
	return TRUE;
}

/***********************************************************************
GPL_Boolean DNA_TestClient::ShutdownHandler::HandleCommand(void)
***********************************************************************/
GPL_Boolean DNA_TestClient::ShutdownHandler::HandleCommand(void)
{
	((DNA_TestClient *) WBA_Application::GetCurrentApplication())->RequestShutdown();
	return TRUE;
}

/******************************************************************************
int main(const int argc, const char **argv, const char **envp)
******************************************************************************/
int main(const int argc, const char **argv, const char **envp)
{
	DNA_TestClient client;

	// Set up the aliases a la getopt(3C).
	WBA_ARGS.Set("DNA_TEST_DEBUG", "0", "Print debugging information", 1, "d");
	WBA_ARGS.Set("DNA_TEST_PRINT_HELP", "0", "Print help and exit", 1, "h");
	WBA_ARGS.Set("DNA_TEST_PORT", "7103", "Server port", 1, "p");
	GPL_String server;
	DNA_IPPoint::GetHost(server);
	WBA_ARGS.Set("DNA_TEST_SERVER", server, "Host name of test server", 1, "s");
	WBA_ARGS.Set("DNA_TEST_PRINT_VERSION", "0", "Print version and exit", 1, "v");

	// Provide scoping for aware objects.
	{
		GPL_Aware gplAware;
		WBA_Aware wbaAware;
		DNA_Aware dnaAware;

		// Load only relevant environment variables.
		WBA_ENV.LoadEnvVariables(envp, FALSE);
	}

	// Load all arguments.
	if (!WBA_ENV.LoadCommandLine(argc, argv, TRUE))
	{
		client.Printf("%s", client.GetUsage());
		client.SetExitCode(-1);
		client.Exit();
	}
	if (WBA_ARGS.Find("DNA_TEST_PRINT_HELP")->IsTrue())
	{
		client.Printf("%s", client.GetUsage());
		client.SetExitCode(0);
		client.Exit();
	}
	if (WBA_ARGS.Find("DNA_TEST_PRINT_VERSION")->IsTrue())
	{
		client.Printf("%s %s", client.GetName(),
						client.GetVersion().ToString().GetData());
		client.SetExitCode(0);
		client.Exit();
	}

	return client.Main();
}

